#include <iostream>

#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_cms.h"
#include "support/configdb.h"
#include "support/tiffsave.h"
#include "support/pathsupport.h"
#include "support/progress.h"
#include "profilemanager/profilemanager.h"
#include "util.h"

#include "it_temp.h"
#include "it_core.h"

#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)

using namespace std;


IT_Settings::IT_Settings(ConfigFile *myfile) : ConfigDB(Template)
{
	new ConfigDBHandler(myfile,"[ITSettings]",this);
}

IT_Settings::~IT_Settings()
{
}


ConfigTemplate IT_Settings::Template[]=
{
	ConfigTemplate("ArgyllPath_TIFFGamut","/usr/local/bin/tiffgamut"),
	ConfigTemplate("ArgyllPath_ICCLink","/usr/local/bin/collink"),
	ConfigTemplate("SourceVC","md"),
	ConfigTemplate("DestVC","pp"),
	ConfigTemplate("GamutMode","-qm -G"),
	ConfigTemplate("SrcProfile",""),
	ConfigTemplate("DstProfile",""),
	ConfigTemplate("Blur",0),
	ConfigTemplate("BlurRadius",15.0),
	ConfigTemplate("Downsample",0),
	ConfigTemplate("DownsamplePercent",50.0),
	ConfigTemplate("GamutReduce",0),
	ConfigTemplate("GamutReducePercent",90),
	ConfigTemplate() // NULL terminated...
};


IT_Core::IT_Core() : ConfigFile(), ProfileManager(this,"[ColourMangement]"), TempFileTracker(), settings(this),
	configfilename(NULL)
{
	// Default preset filename
	configfilename=substitute_homedir("$HOME/.imagetargetrc");
}


IT_Core::~IT_Core()
{
	if(configfilename)
		free(configfilename);
}


void IT_Core::ProcessFile(const char *filename,Progress *p)
{
	// Save a temp copy of the image
	Temp_SrcImage si(this,filename,p);

	// And create a tailored Devicelink
	Temp_DeviceLink tdl(this,&si,p);

	// Create transform from DeviceLink profile, and transform image.
	CMSProfile dl(tdl.Filename());
	if(!dl.IsDeviceLink())
		throw _("The profile we just created doesn't seem to be a device link.");
	CMSTransform trans(&dl);

	// Now load the image
	ImageSource *is=ISLoadImage(si.Filename());

	is=new ImageSource_CMS(is,&trans);

	// We embed the provided output file in the targetted TIFF,
	// rather than the DeviceLink!
	CMSProfile *dstp=GetProfile(settings.FindString("DstProfile"));
	is->SetEmbeddedProfile(dstp);

	// Now we save the Targetted TIFF.
	char *outname=BuildFilename(filename,"-targetted","tif");

	TIFFSaver ts(outname,is,16);
	ts.SetProgress(p);
	ts.Save();
	delete is;
	delete dstp;
	free(outname);
}


void IT_Core::ParseOptions(int argc,char *argv[])
{
	static struct option long_options[] =
	{
		{"help",no_argument,NULL,'h'},
		{"version",no_argument,NULL,'v'},
		{"tiffgamut-path",required_argument,NULL,'t'},
		{"icclink-path",required_argument,NULL,'l'},
		{"default-profile",required_argument,NULL,'p'},
		{"output-profile",required_argument,NULL,'o'},
		{"sourcevc",required_argument,NULL,'s'},
		{"destvc",required_argument,NULL,'d'},
		{0, 0, 0, 0}
	};

	while(1)
	{
		int c;
		c = getopt_long(argc,argv,"hvt:l:p:o:s:d:",long_options,NULL);
		if(c==-1)
			break;
		switch (c)
		{
			case 'v':
				cerr << PACKAGE_STRING << endl;
				throw 0;
				break;
			case 't':
				settings.SetString("ArgyllPath_TIFFGamut",optarg);
				break;
			case 'l':
				settings.SetString("ArgyllPath_ICCLink",optarg);
				break;
			case 'p':
				{
					CMSProfile *p=GetProfile(optarg);
					if(p)
						settings.SetString("SrcProfile",optarg);
					delete p;
				}
				break;
			case 'o':
				{
					CMSProfile *p=GetProfile(optarg);
					if(p)
						settings.SetString("DstProfile",optarg);
					delete p;
				}
				break;
			case 's':
				settings.SetString("SourceVC",optarg);
				break;
			case 'd':
				settings.SetString("DestVC",optarg);
				break;
			case 'h':
			default:
				cerr << "Usage: " << argv[0] << "[options] image1 ... " << endl;
				cerr << "\t-h --help\t\tdisplay this message" << endl;
				cerr << "\t-v --version\t\tdisplay version" << endl;
				cerr << "\t-t --tiffgamut-path\tPath to Argyll TIFFGamut utility" << endl;
				cerr << "\t-l --icclink-path\tPath to Argyll ICCLink utility" << endl;
				cerr << "\t-p --default-profile\tUse the specified ICC profile"
					<< endl << "\t\t\t\tif no embedded profile is present" << endl;
				cerr << "\t-o --output-profile\tICC profile of output device" << endl;
				cerr << "\t-s --sourcevc\t\tSource viewing conditions (Argyll two-letter code)" << endl;
				cerr << "\t-d --destvc\t\tDestination viewing conditions (Argyll two-letter code)" << endl;
				// FIXME - enumerate viewing condition codes here...
				cerr << endl << "An output TIFF file will be created, transformed to the" << endl;
				cerr << "output profile's colour space with a transform tailored to the" << endl;
				cerr << "specified image, using Argyll's tiffgamut and icclink utilities." << endl;
				throw 0;
				break;
		}
	}
}


void IT_Core::ParseConfigFile()
{
	ConfigFile::ParseConfigFile(configfilename);
	// Having parsed the config file, we copy the Default RGB, and Printer profiles
	// to IT_Core's own settings.  They can still be overridden by the command
	// line arguments.

	settings.SetString("SrcProfile",FindString("DefaultRGBProfile"));
	settings.SetString("DstProfile",FindString("PrinterProfile"));
}


bool IT_Core::SaveConfigFile()
{
	return(ConfigFile::SaveConfigFile(configfilename));
}


bool IT_Core::CheckArgyllUtils()
{
	const char *tgpath=settings.FindString("ArgyllPath_TIFFGamut");
	const char *icpath=settings.FindString("ArgyllPath_ICCLink");
	FILE *f;
	if(!(f=fopen(tgpath,"r")))
		return(false);
	fclose(f);
	if(!(f=fopen(icpath,"r")))
		return(false);
	fclose(f);
	
	return(true);
}

