#include <iostream>
#include <fstream>

#include "config.h"

#include "util.h"
#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_cms.h"
#include "imagesource/imagesource_gaussianblur.h"
#include "imagesource/imagesource_downsample.h"
#include "support/tiffsave.h"
#include "config.h"


#include "it_core.h"
#include "it_temp.h"

#undef INT32
#include "thread.h"

#include "gettext.h"
#define _(x) gettext(x)

using namespace std;


// Source image tempfile.


Temp_SrcImage::Temp_SrcImage(IT_Core *header,const char *srcfn,Progress *p)
	: TempFile(header,"imtg"), srcprofile(NULL), gamut(NULL), progress(p)
{
	char *tmp=tempnam(NULL,prefix);
	filename=BuildFilename(tmp,"_src.tif","");
	free(tmp);

	ImageSource *is=ISLoadImage(srcfn);
	if(is)
	{
		// Save off source profile
		
		srcprofile=new Temp_SrcProfile(header,is);

		if(header->settings.FindInt("Blur"))
			is=new ImageSource_GaussianBlur(is,header->settings.FindFloat("BlurRadius"));
		
		if(header->settings.FindInt("Downsample"))
		{
			float p=0.01*header->settings.FindFloat("DownsamplePercent");
			is=new ImageSource_Downsample(is,p*is->width,p*is->height);
		}

		progress->SetMessage(_("Saving temporary copy of image"));
		// Copy the image to a temporary file;
		TIFFSaver *ts=new TIFFSaver(Filename(),is,16);
		ts->SetProgress(progress);
		ts->Save();
		delete ts;
		delete is;
	}
}

Temp_SrcImage::~Temp_SrcImage()
{
	if(srcprofile)
		delete srcprofile;
	if(gamut)
		delete gamut;
}


Temp_SrcProfile *Temp_SrcImage::GetSrcProfile()
{
	if(!srcprofile)
		srcprofile=new Temp_SrcProfile((IT_Core *)header,NULL);
	return(srcprofile);
}


Temp_Gamut *Temp_SrcImage::GetGamut()
{
	if(!gamut)
		gamut=new Temp_Gamut((IT_Core *)header,this,progress);
	return(gamut);
}


void Temp_SrcImage::FlushGamut()
{
	if(gamut)
		delete gamut;
	gamut=NULL;
}


void Temp_SrcImage::FlushSrcProfile()
{
	if(srcprofile)
		delete(srcprofile);
	srcprofile=NULL;
}


// Source ICC Profile tempfile


Temp_SrcProfile::Temp_SrcProfile(IT_Core *header,ImageSource *is) : TempFile(header,"imtg")
{
	char *tmp=tempnam(NULL,prefix);
	filename=BuildFilename(tmp,"_src.icc","");
	free(tmp);

	// Determine a suitable ICC profile for the image.
	// We use the embedded profile if present, and a
	// default profile if not.
	CMSProfile *prof=NULL;
	if(is)
		prof=is->GetEmbeddedProfile();
	if(!prof)
	{
		prof=header->GetProfile(header->settings.FindString("SrcProfile"));
	}
	if(!prof)
		throw "Can't open source profile";
	cerr << "Using profile " << *prof << endl;
	
	// Save profile to a temporary file.
	cerr << "Saving source profile to: " << Filename() << endl;
	prof->Save(Filename());
	// Free the profile if it wasn't embedded.
	if(is && !is->GetEmbeddedProfile())
		delete prof;
}


Temp_SrcProfile::~Temp_SrcProfile()
{
}


// Destination ICC Profile tempfile
// We copy the profile to tempfile with a name of our choosing
// to sidestep the issue of escaping the filename for the command-line.


Temp_DestProfile::Temp_DestProfile(IT_Core *header,const char *pfn) : TempFile(header,"imtg")
{
	char *tmp=tempnam(NULL,prefix);
	filename=BuildFilename(tmp,"_dst.icc","");
	free(tmp);

	if(pfn)
	{
#define TMPBUFSIZE 4096
		char buffer[TMPBUFSIZE];
		ifstream in(pfn,ios::in|ios::binary);
		ofstream out(filename,ios::out|ios::binary);
		while(!in.eof())
		{
			in.read(buffer,TMPBUFSIZE);
			int n=in.gcount();
			if(n)
				out.write(buffer,n);
		}
		out.close();
		in.close();
	}
}


Temp_DestProfile::~Temp_DestProfile()
{
}


// Shared function to execute a system command from within a thread
#if defined HAVE_LIBPTHREAD || defined HAVE_LIBPTHREADGC2

static int thread_entry(Thread *t,void *ud)
{
	int returncode=0;
	const char *cmd=(const char *)ud;
	t->SendSync();
	try
	{
		cerr << "Calling command:..." << cmd << "..." << endl;
		if(system(cmd))
			throw "Devicelink generation failed";
	}
	catch(const char *err)
	{
		cerr << "Subthread error: " << err << endl;
		returncode=-1;
	}
	return(returncode);
}

#endif

// Gamut tempfile


Temp_Gamut::Temp_Gamut(IT_Core *header,Temp_SrcImage *img, Progress *prog) : TempFile(header,"imtg")
{
	filename=BuildFilename(img->Filename(),"","gam");

	Temp_SrcProfile *prof=img->GetSrcProfile();

	char *filter=(char *)"";
	if(header->settings.FindInt("GamutReduce"))
	{
		filter=(char *)malloc(10);
		sprintf(filter,"-f %d",header->settings.FindInt("GamutReducePercent"));
	}

	// Generate a command line for tiffgamut
	const char *tg_cmdfmt="%s -ir -pj %s -c%s %s %s";
	const char *argyll_path=header->settings.FindString("ArgyllPath_TIFFGamut");
	const char *source_vc=header->settings.FindString("SourceVC");

	int tg_cmdlen=strlen(tg_cmdfmt) + strlen(argyll_path) + strlen(source_vc) + strlen(filter)
		+ strlen(prof->Filename()) + strlen(img->Filename()) + 4;
		
	cerr << "Allocating " << tg_cmdlen << " bytes for tiffgamut command" << endl;

	char *tg_cmd=(char *)malloc(tg_cmdlen);
	sprintf(tg_cmd,tg_cmdfmt,argyll_path,filter,source_vc,prof->Filename(),img->Filename());

	if(strlen(filter))
		free(filter);

	prog->SetMessage(_("Analyzing image gamut\nplease wait..."));
	prog->DoProgress(0,0);
#if defined HAVE_LIBPTHREAD || defined HAVE_LIBPTHREADGC2
	Thread t(thread_entry,tg_cmd);
	t.Start();
	t.WaitSync();
	while(!t.TestFinished())
	{
		prog->DoProgress(0,0);
#ifdef WIN32
		Sleep(50);
#else
		usleep(50000);
#endif
	}
	if(t.GetReturnCode()!=0)
		throw _("Gamut generation failed");
#else
	if(system(tg_cmd))
		throw _("Gamut generation failed");
#endif
	free(tg_cmd);
}


Temp_Gamut::~Temp_Gamut()
{
}


// DeviceLink tempfile


Temp_DeviceLink::Temp_DeviceLink(IT_Core *header,Temp_SrcImage *img, Progress *prog) : TempFile(header,"imtg")
{
	filename=BuildFilename(img->Filename(),"_dl","icc");

	Temp_SrcProfile *prof=img->GetSrcProfile();
	Temp_Gamut *gamut=img->GetGamut();
	CMSProfile *destprofile=header->GetProfile(header->settings.FindString("DstProfile"));
	if(!destprofile)
		throw _("Please choose a destination profile!");

	// Copy the output profile to a tempfile to sidestep the issue of escaping filename
	// for the command line.
	Temp_DestProfile tdp(header,destprofile->GetFilename());

	// Generate a command line for icclink
	const char *icl_cmdfmt="%s %s %s -ip -op -ip -c%s -d%s %s %s %s";
	const char *argyll_path=header->settings.FindString("ArgyllPath_ICCLink");
	const char *source_vc=header->settings.FindString("SourceVC");
	const char *dest_vc=header->settings.FindString("DestVC");
	const char *gamutmode=header->settings.FindString("GamutMode");

	int icl_cmdlen=strlen(icl_cmdfmt) + strlen(argyll_path) + strlen(gamut->Filename())
		+ strlen(source_vc) + strlen(dest_vc)
		+ strlen(prof->Filename()) + strlen(tdp.Filename()) + strlen(Filename()) + 4;

	cerr << "Allocating " << icl_cmdlen << " bytes for icclink command" << endl;

	char *icl_cmd=(char *)malloc(icl_cmdlen);
	sprintf(icl_cmd,icl_cmdfmt,argyll_path,gamutmode,gamut->Filename(),source_vc,dest_vc,
		prof->Filename(),tdp.Filename(),Filename());

	delete destprofile;

	prog->SetMessage(_("Building custom devicelink profile.\nThis can take a while -\nPlease be patient!"));
	prog->DoProgress(0,0);

#if defined HAVE_LIBPTHREAD || defined HAVE_LIBPTHREADGC2
	Thread t(thread_entry,icl_cmd);
	t.Start();
	t.WaitSync();
	prog->SetMessage(_("Building custom devicelink profile.\nThis can take a while -\nPlease be patient!"));
	while(!t.TestFinished())
	{
		prog->DoProgress(0,0);
#ifdef WIN32
		Sleep(50);
#else
		usleep(50000);
#endif
	}
	if(t.GetReturnCode()!=0)
		throw _("DeviceLink generation failed");
#else
	if(system(icl_cmd))
		throw _("DeviceLink generation failed");
#endif
	free(icl_cmd);
}


Temp_DeviceLink::~Temp_DeviceLink()
{
}


// Temp transformed image


Temp_Image::Temp_Image(IT_Core *header,const char *fn,Temp_DeviceLink *tdl,Progress *p) : TempFile(header,"imtg")
{
	char *tmp=tempnam(NULL,prefix);
	filename=BuildFilename(tmp,"_out.tif","");
	free(tmp);

	// Create transform from DeviceLink profile, and transform image.
	CMSProfile dl(tdl->Filename());
	if(!dl.IsDeviceLink())
		throw "The profile we just created doesn't seem to be a device link.";
	CMSTransform trans(&dl);

	// Now load the image
	ImageSource *is=ISLoadImage(fn);

	is=new ImageSource_CMS(is,&trans);

	// We embed the provided output file in the targetted TIFF,
	// rather than the DeviceLink!
	CMSProfile *dstp=header->GetProfile(header->settings.FindString("DstProfile"));
	is->SetEmbeddedProfile(dstp);

	p->SetMessage(_("Applying DeviceLink to image"));
	TIFFSaver ts(Filename(),is,16);
	ts.SetProgress(p);
	ts.Save();
	delete is;
	delete dstp;
}

Temp_Image::~Temp_Image()
{
}
